"""
This module exists to prevent having to use `eval()`.

`ast.literal_eval()` is not a drop-in replacement however, the function `safe_eval.safe_eval()` will catch some edge cases.

A downside of literal_eval() is that is cannot parse
special characters like newlines (\r\t or \n). We need those kind of characters because pandas.read_csv() accepts 
a parameter 'sep' that could contain all sorts of regex.

As an example, if we have this in our markdown file:

```markdown
{{ read_csv('my/path/table.csv', sep = '\t\n') }}
```

We use regex to first extract the argkwarg string:

"'my/path/table.csv', sep = '\t\n'"

And we then need to parse that into args and kwargs:

>>> args
['my/path/table.csv']
>>> kwargs
{'sep' : '\t\n'}

So we can finally use those to safely run pd.read_csv(*args, **kwargs)

"""

import re
from ast import literal_eval


def safe_eval(string):
    """
    A downside of literal_eval() is that is cannot parse
    special characters like newlines (\r\t or \n).
    
    We need this because pandas.read_csv() accepts 
    a parameter 'sep' that could contain all sorts of regex.

    Args:
        string (str): string to parse to literal python

    Returns:
        str: The parsed literal python structure 
    """
    if "\n" in string or "\\" in string or "\r" in string:
        # remove quotes
        string = string.replace("'", "")
        string = string.replace('"', "")
        return string
    else:
        return literal_eval(string)


def parse_argkwarg(string: str):
    """
    Parses a string to detect both args and kwargs.
    
    Adapted code from 
    https://stackoverflow.com/questions/9305387/string-of-kwargs-to-kwargs

    Args:
        string (str): string with positional and keyword arguments
        
    Returns:
        args[List], kwargs[Dict]
    """

    argkwargs = re.split(r"(?<!\=)(?:,{1} )(?!\=)", string)

    args = []
    kwargs = []

    for i in argkwargs:
        i = i.strip()
        if "=" in i:
            kwargs.append(i)
        else:
            if len(kwargs) != 0:
                raise AssertionError(
                    "[table-reader-plugin] Make sure the python in your reader tag is correct: Positional arguments follow keyword arguments in '%s'"
                    % string
                )
            args.append(literal_eval(i))

    # kwargs as dict
    kwargs = [re.split(" ?= ?", x) for x in kwargs]
    kwargs = dict([(x[0], safe_eval(x[1])) for x in kwargs])

    return args, kwargs



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mkdocs-table-reader-plugin",
    version="0.6.2",
    description="MkDocs plugin to directly insert tables from files into markdown.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="mkdocs plugin",
    url="https://github.com/timvink/mkdocs-table-reader-plugin",
    author="Tim Vink",
    author_email="vinktim@gmail.com",
    license="MIT",
    python_requires=">=3.6",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=["mkdocs>=1.0", "pandas>=1.1", "tabulate>=0.8.7", "PyYAML>=5.4.1"],
    packages=find_packages(),
    entry_points={
        "mkdocs.plugins": [
            "table-reader = mkdocs_table_reader_plugin.plugin:TableReaderPlugin"
        ]
    },
)



import os
import re
import pandas as pd
import yaml
import textwrap

from mkdocs.plugins import BasePlugin
from mkdocs.config import config_options
from mkdocs.exceptions import ConfigurationError

from .safe_eval import parse_argkwarg


def read_csv(*args, **kwargs):
    df = pd.read_csv(*args, **kwargs)
    return df.to_markdown(index=False, tablefmt="pipe")


def read_table(*args, **kwargs):
    df = pd.read_table(*args, **kwargs)
    return df.to_markdown(index=False, tablefmt="pipe")


def read_fwf(*args, **kwargs):
    df = pd.read_fwf(*args, **kwargs)
    return df.to_markdown(index=False, tablefmt="pipe")


def read_excel(*args, **kwargs):
    df = pd.read_excel(*args, **kwargs)
    return df.to_markdown(index=False, tablefmt="pipe")


def read_yaml(*args, **kwargs):
    with open(args[0], "r") as f:
        df = pd.json_normalize(yaml.safe_load(f), **kwargs)

    return df.to_markdown(index=False, tablefmt="pipe")


READERS = {
    "read_csv": read_csv,
    "read_table": read_table,
    "read_fwf": read_fwf,
    "read_excel": read_excel,
    "read_yaml": read_yaml,
}


class cd:
    """
    Context manager for changing the current working directory
    Credits: https://stackoverflow.com/a/13197763/5525118
    """

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


class TableReaderPlugin(BasePlugin):

    config_scheme = (("data_path", config_options.Type(str, default=".")),)

    def on_config(self, config):
        """

        See https://www.mkdocs.org/user-guide/plugins/#on_config
        Args:
            config

        Returns:
            Config
        """

        plugins = [p for p in config.get("plugins")]

        for post_load_plugin in ["macros", "markdownextradata"]:
            if post_load_plugin in plugins:
                if plugins.index("table-reader") > plugins.index(post_load_plugin):
                    raise ConfigurationError(f"[table-reader]: Incompatible plugin order: Define 'table-reader' before '{post_load_plugin}' in your mkdocs.yml.")


    def on_page_markdown(self, markdown, page, config, files, **kwargs):
        """
        Replace jinja tag {{ read_csv() }} in markdown with markdown table.

        The page_markdown event is called after the page's markdown is loaded
        from file and can be used to alter the Markdown source text.
        The meta- data has been stripped off and is available as page.meta
        at this point.

        https://www.mkdocs.org/user-guide/plugins/#on_page_markdown

        Args:
            markdown (str): Markdown source text of page as string
            page: mkdocs.nav.Page instance
            config: global configuration object
            site_navigation: global navigation object

        Returns:
            str: Markdown source text of page as string
        """

        mkdocs_dir = os.path.dirname(os.path.abspath(config["config_file_path"]))

        for reader, function in READERS.items():
            
            # Regex pattern for tags like {{ read_csv(..) }}
            # match group 0: to extract any leading whitespace 
            # match group 1: to extract the arguments (positional and keywords)
            tag_pattern = re.compile(
                "( *)\{\{\s+%s\((.+)\)\s+\}\}" % reader, flags=re.IGNORECASE
            )

            matches = re.findall(tag_pattern, markdown)
            
            
            for result in matches:

                # Safely parse the arguments
                pd_args, pd_kwargs = parse_argkwarg(result[1])

                # Make sure the path is relative to "data_path"
                if len(pd_args) > 0:
                    pd_args[0] = os.path.join(self.config.get("data_path"), pd_args[0])
                    file_path = pd_args[0]

                if pd_kwargs.get("filepath_or_buffer"):
                    file_path = pd_kwargs["filepath_or_buffer"]
                    file_path = os.path.join(self.config.get("data_path"), file_path)
                    pd_kwargs["filepath_or_buffer"] = file_path

                # Load the table
                with cd(mkdocs_dir):
                    if not os.path.exists(file_path):
                        raise FileNotFoundError(
                            "[table-reader-plugin]: File does not exist: %s" % file_path
                        )

                    markdown_table = function(*pd_args, **pd_kwargs)

                # Deal with indentation
                # f.e. relevant when used inside content tabs
                leading_spaces = result[0]
                # make sure it's in multiples of 4 spaces
                leading_spaces = int(len(leading_spaces) / 4) * "    "
                # indent entire table
                fixed_lines = []
                for line in markdown_table.split('\n'):
                    fixed_lines.append(textwrap.indent(line, leading_spaces))
                
                markdown_table = "\n".join(fixed_lines)

                # Insert markdown table
                # By replacing first occurance of the regex pattern
                markdown = tag_pattern.sub(markdown_table, markdown, count=1)


        return markdown
